FORCE_INFINITE_AMMO = true

allSounds = {}
bulletsInAir = {}
bulletType = {
	["damage"] = 1.2,
	["constant_velocity"] = 400
}
gunMode = 0

GUN_INACTIVE = 0
GUN_CHARGING_UP = 1
GUN_READY = 3
GUN_FIRING = 4

BULLET_FIRE_LOCATION = Vec(16.1, 64.3, -9.0)
TIME_TO_CHARGE_UP = 1.4
MAX_RECOIL = 0.55
VOXELS_PER_AMMO = 20000

readyTimer = 0
cooldownTimer = 0
fovModifier = 0
desiredEmissive = 0
recoilStrength = 0
currentBrokenVoxelCount = 0
lastBrokenVoxelCount = 0
bulletFireLocation = nil

function CopyTable(orig)
    local orig_type = type(orig)
    local copy = nil

    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[CopyTable(orig_key)] = CopyTable(orig_value)
        end
        setmetatable(copy, CopyTable(getmetatable(orig)))
    else
        copy = orig
    end

    return copy
end

function TableSerialize(inp)
    if type(inp) ~= "table" then return tostring(inp) end

    local res = "{"
    for k, v in pairs(inp) do
        if type(k) == "table" then
            k = TableSerialize(k)
        elseif type(k) ~= "number" then
            k = '"' .. k .. '"'
        end
        res = res .. "[" .. k .. "] = " .. TableSerialize(v) .. ", "
    end

    return res:sub(1, #res - 2) .. "}"
end

OldDebugPrint = DebugPrint
function DebugPrint(o)
    OldDebugPrint(TableSerialize(o))
end

function MagicaVoxelPointToLocalPoint(vec)
	-- X Z -Y, VOX_SIZE = 0.05
	return VecScale(Vec(vec[1], vec[3], -vec[2]), 0.05)
end

function NumEqualsWithTolerance(val1, val2, tolerance)
	return math.abs(val1 - val2) <= tolerance
end

function VecEqualsWithTolerance(vec1, vec2, tolerance)
	return NumEqualsWithTolerance(vec1[1], vec2[1], tolerance) and NumEqualsWithTolerance(vec1[2], vec2[2], tolerance) and NumEqualsWithTolerance(vec1[3], vec2[3], tolerance)
end

function VecDist(vec1, vec2)
	return VecLength(VecSub(vec1, vec2))
end

function VecDist2D(vec1, vec2)
	vec1H = VecCopy(vec1)
	vec1H[2] = 0
	vec2H = VecCopy(vec2)
	vec2H[2] = 0
	return VecDist(vec1H, vec2H)
end

function RandomVector3D(length)
	local theta = math.random() * 2 * math.pi
	local a = math.random() * 2 - 1
	local b = math.sqrt(1 - math.pow(a, 2))
	return VecScale(Vec(b * math.cos(theta), b * math.sin(theta), a), length)
end

function math.clamp(val, lower, upper)
    if lower > upper then lower, upper = upper, lower end -- swap if boundaries supplied the wrong way
    return math.max(lower, math.min(upper, val))
end

function ShootArmageddon()	
	local ct = GetCameraTransform()
	local forwardPoint = TransformToParentPoint(ct, Vec(0, 0, -1))
	local relativeDirection = VecNormalize(VecSub(forwardPoint, ct.pos))
	local relativeLength = VecDist(forwardPoint, ct.pos)
	local hit, dist = QueryRaycast(ct.pos, relativeDirection, relativeLength * 100)

	PlaySound(allSounds["fire"], ct.pos, 1.0)

	local bulletDirection = VecNormalize(VecSub(VecAdd(ct.pos, VecScale(relativeDirection, dist)), bulletFireLocation))
	local nextBullet = CopyTable(bulletType)
	nextBullet.isInMotion = true
	nextBullet.isDead = false
	nextBullet.age = 0
	nextBullet.pos = VecCopy(bulletFireLocation)
	nextBullet.dir = VecCopy(bulletDirection)
	nextBullet.velocity = VecScale(nextBullet.dir, bulletType["constant_velocity"] or (dist / bulletType["time_to_hit_target"]))
	bulletsInAir[#bulletsInAir + 1] = nextBullet	

	SetInt("game.tool.armageddon.ammo", GetInt("game.tool.armageddon.ammo") - 1)
	SetValue("recoilStrength", MAX_RECOIL, "easeout", 0.15)
	cooldownTimer = 1
end

function ProjectileOperations(bullet)
	-- We need to check: does something intercept the path of our bullet this tick?
	local newPosition = VecAdd(bullet.pos, VecScale(bullet.velocity, GetTimeStep()))
	local relativeBullet = VecSub(newPosition, bullet.pos)
	local distTravelling = VecLength(relativeBullet)
	local direction = VecNormalize(relativeBullet)
	QueryRequire("physical")
	local hit, dist = QueryRaycast(bullet.pos, direction, distTravelling)
	
	if hit then
		hitPos = VecAdd(bullet.pos, VecScale(direction, dist))
		for i = 1, 2 do
			MakeHole(hitPos, bullet.damage, bullet.damage, bullet.damage)
		end
	end

	local lerpNow = 0
	while lerpNow < 1 do
		lerpNow = lerpNow + 0.1
		hitPos = VecLerp(bullet.pos, newPosition, lerpNow)
		for i = 1, 2 do
			MakeHole(hitPos, bullet.damage, bullet.damage, bullet.damage)
		end
	end

	-- Send stuff flying
	local max = bullet.damage * 0.8
	local allBodiesToFling = {}
	local hash = {}
	local lerpNow = 0
	while lerpNow < 1 do
		lerpNow = lerpNow + 0.1
		local hitPos = VecLerp(bullet.pos, newPosition, lerpNow)
		QueryRequire("physical dynamic")
		local allCurrentBodiesToFling = QueryAabbBodies(VecAdd(hitPos, Vec(-max, -max, -max)), VecAdd(hitPos, Vec(max, max, max)))
		for i = 1, #allCurrentBodiesToFling do
			if not hash[allCurrentBodiesToFling[i]] then
				allBodiesToFling[#allBodiesToFling + 1] = allCurrentBodiesToFling[i]
				hash[allCurrentBodiesToFling[i]] = true
			end
		end
	end

	for i = 1, #allBodiesToFling do
		local thisBodyToFling = allBodiesToFling[i]
		local applyToSpeed = math.random(15, 25)
		applyToSpeed = math.clamp(applyToSpeed * (GetBodyMass(thisBodyToFling) / 5000), applyToSpeed * 0.5, applyToSpeed * 1.5)
		local applyToVelocity = VecAdd(VecScale(direction, applyToSpeed), RandomVector3D(applyToSpeed * 0.5))
		SetBodyVelocity(thisBodyToFling, VecAdd(GetBodyVelocity(thisBodyToFling), applyToVelocity))
	end

	-- We need to die if we're too old
	bullet.age = bullet.age + GetTimeStep()
	if bullet.age > 5 then
		bullet.isInMotion = false
		bullet.isDead = true
	end
	
	local distBetweenParticles = bullet.damage * 0.9
	local distNow = 0
	while distNow < distTravelling do
		ParticleReset()
		ParticleType("plain")
		ParticleTile(5)
		ParticleColor(1, 0.65, 0)
		ParticleRadius(bullet.damage * 0.8)
		ParticleAlpha(1, 0.25)
		ParticleGravity(0)
		ParticleDrag(0)
		ParticleEmissive(2, 0.1)
		ParticleStretch(0)
		ParticleSticky(0)
		ParticleCollide(0)
		SpawnParticle(VecAdd(bullet.pos, VecScale(direction, distNow)), Vec(0,0,0), 0.5)
		distNow = distNow + distBetweenParticles
	end

	bullet.pos = newPosition
end

function init()
	RegisterTool("armageddon", "ARMAGEDDON", "MOD/vox/ARMAGEDDON.vox")
	SetBool("game.tool.armageddon.enabled", true)
	SetInt("game.tool.armageddon.ammo", 30)
	
	allSounds["charge"] = LoadLoop("MOD/snd/ARMAGEDDON_charge_loop.ogg")
	allSounds["overcharge"] = LoadLoop("MOD/snd/ARMAGEDDON_overcharge_loop.ogg")
	allSounds["fire"] = LoadSound("MOD/snd/ARMAGEDDON_firing.ogg")
	allSounds["ready"] = LoadSound("MOD/snd/ARMAGEDDON_ready.ogg")
end

-- Updates once per frame
function tick(dt)
	if GetBool("savegame.mod.armageddon.infiniteAmmo") or GetBool("level.unlimitedammo") then 
		SetInt("game.tool.armageddon.ammo", 9999)
	else
		currentBrokenVoxelCount = currentBrokenVoxelCount + (GetInt("game.brokenvoxels") - lastBrokenVoxelCount)
		lastBrokenVoxelCount = GetInt("game.brokenvoxels")

		if currentBrokenVoxelCount >= VOXELS_PER_AMMO then
			SetInt("game.tool.armageddon.ammo", GetInt("game.tool.armageddon.ammo") + 1)
			currentBrokenVoxelCount = 0
		end
	end

	-- Tick active bullets
	local cleaningUpBullets = {}

	for i = 1, #bulletsInAir do
		local thisBullet = bulletsInAir[i]
		if thisBullet == nil or thisBullet.isDead then
			cleaningUpBullets[#cleaningUpBullets + 1] = i
		elseif thisBullet.isInMotion then
			ProjectileOperations(thisBullet)
		end
	end

	cleanedBulletsInAir = {}
	for i = 1, #cleaningUpBullets do
		bulletsInAir[cleaningUpBullets[i]] = nil
	end
	for i = 1, #bulletsInAir do
		if bulletsInAir[i] ~= nil then cleanedBulletsInAir[#cleanedBulletsInAir + 1] = bulletsInAir[i] end
	end
	bulletsInAir = cleanedBulletsInAir

	if not GetBool("savegame.mod.armageddon.noFovAdjustment") then
		SetCameraFov(GetInt("options.gfx.fov") - fovModifier)
	end

	if GetString("game.player.tool") == "armageddon" and GetBool("game.player.canusetool") and GetFloat("game.player.health") ~= 0 then
		-- Charging up
		if gunMode == GUN_READY then
			if not InputDown("usetool") then
				gunMode = GUN_FIRING
				SetValue("fovModifier", 10, "linear", 0.08)
			end

			if readyTimer >= 0.5 then
				PlayLoop(allSounds["overcharge"], GetPlayerTransform().pos, 0.65)
			end
		else
			if GetInt("game.tool.armageddon.ammo") > 0 and cooldownTimer <= 0 and recoilStrength == 0 and InputDown("usetool") then
				PlayLoop(allSounds["charge"], GetPlayerTransform().pos, math.clamp(desiredEmissive * 0.75, 0, 0.9))
				gunMode = GUN_CHARGING_UP
				SetValue("desiredEmissive", 1.0, "linear", (1.0 - desiredEmissive) * TIME_TO_CHARGE_UP)
			else
				gunMode = GUN_INACTIVE
				SetValue("desiredEmissive", 0.0, "linear", desiredEmissive * TIME_TO_CHARGE_UP)
			end
		end

		-- FOV
		if fovModifier >= 10 then
			SetValue("fovModifier", 0, "linear", 0.2)
		end

		-- Recoil
		if recoilStrength >= MAX_RECOIL then
			SetValue("recoilStrength", 0.0, "easein", 0.5)
		end

		local toolBody = GetToolBody()
		if toolBody ~= 0 then
			SetToolTransform(Transform(Vec(0, 0, recoilStrength)))
		end
	else
		gunMode = GUN_INACTIVE
		desiredEmissive = 0
	end
end

-- Updates always 60 times per second
function update(dt)
	if cooldownTimer > 0 then
		cooldownTimer = cooldownTimer - dt
	end

	if gunMode == GUN_CHARGING_UP then
		if desiredEmissive >= 1.0 then
			PlaySound(allSounds["ready"], GetPlayerTransform().pos, 0.65)
			gunMode = GUN_READY
		end
	end

	if gunMode == GUN_READY then
		readyTimer = readyTimer + dt
	else
		readyTimer = 0
	end

	local toolBody = GetToolBody()
	if GetString("game.player.tool") == "armageddon" and toolBody ~= 0 then
		-- Get some relative positions based off of the tool body
		local toolTrans = GetBodyTransform(toolBody)
		bulletFireLocation = TransformToParentPoint(toolTrans, MagicaVoxelPointToLocalPoint(BULLET_FIRE_LOCATION))

		local allToolShapes = GetBodyShapes(toolBody)
		for i = 1, #allToolShapes do
			local thisToolShape = allToolShapes[i]
			SetShapeEmissiveScale(thisToolShape, desiredEmissive)
		end

		-- Firing bullets
		if gunMode == GUN_FIRING and cooldownTimer <= 0 then
			ShootArmageddon()
			gunMode = GUN_INACTIVE
		end
	end
end
